#ifndef CGAL_B_SPLINE_CURVE_2_H
#define CGAL_B_SPLINE_CURVE_2_H

#include "Types.h"
#include "Helpers.h"
#include "Curve_constants.h"
#include "Parametric_curve_2.h"
#include "Subdivision_curve_2.h"
#include "DeBoor_FO.h"
#include "X_FO.h"
#include "Y_FO.h"


// Template class B_spline_curve_2 implements a B-Spline curve in 2D
// Template parameters:
//       NumType_      number type for coordinates
//       A_            pointer to function object which implements 
//                     the evaluation of a point on the curve at given
//                     parameter value t
// Inherits from the classes Parametric_curve_2 and Subdivision_curve_2

template <class NumType_, class A_ = DeBoor_FO<NumType_> > 
class B_spline_curve_2 : public Parametric_curve_2<NumType_, 
                                                   X_FO<NumType_, A_*>*, 
		                                   Y_FO<NumType_, A_*>* >, 
			 public Subdivision_curve_2<NumType_> {

public:    
    typedef A_ A;
    typedef X_FO<NumType, A* > B_spline_x;
    typedef Y_FO<NumType, A* > B_spline_y;
    typedef Parametric_curve_2<NumType, B_spline_x*, B_spline_y*> P_curve;
    typedef vector<NumType> ParamVector;    
    typedef Subdivision_curve_2<NumType> S_curve;

    
    // B_spline_curve_2: void -> B_spline_curve_2    
    // Effect: contructs a empty B_spline_curve_2 object whose parameter is
    //         in the default interval [LOW_BOUND, UP_BOUND]
    // contructor without parameter    
    B_spline_curve_2() : S_curve(), degree_(0), knots_(),
	alg(&contr_points, &knots_, degree_), b_x_FO(&alg), b_y_FO(&alg) {
	P_curve::init(&b_x_FO, &b_y_FO, LOW_BOUND, UP_BOUND);
    }

    // B_spline_curve_2: Points, vector<NumType>, unsigned -> 
    //                   B_spline_curve_2 
    // Parameters:
    //    contr_pts  the control points of the B_spline curve
    //    knots      the knots of the B_spline curve
    //    degree     the degree of the B_spline curve, default value 3 
    // Pre:
    //    !contr_pts.empty() AND contr_pts.size() == knots.size()
    // Effect: contructs a B_spline_curve_2 object with control points 
    //         contr_pts of the given degree defined over the knot 
    //         sequence knots
    B_spline_curve_2(const Points & contr_pts, 
		     const ParamVector & knots, unsigned degree = 3) : 
	           S_curve(contr_pts), 
		   degree_(degree), knots_(knots),
	           alg(&contr_points, &knots_, degree_),    	
                   b_x_FO(&alg), b_y_FO(&alg) {	
	assert(!contr_pts.empty() && contr_pts.size()==knots.size());
	assert(contr_pts.size()>degree_);
	P_curve::init(&b_x_FO, &b_y_FO, knots[0], knots[knots.size()-1]);
	initial_approximation();
    }; 

    // B_spline_curve_2: Points, NumType, NumType -> B_spline_curve_2 
    // Parameters:
    //    contr_pts   the control points of the B_spline curve
    //    degree      the degree of the B_spline curve, default value 3
    //    a_          the low bound of t's range, t in [a_,b_], default
    //                value LOW_BOUND
    //    b_          the up bound of t's range, t in [a_,b_], default
    //                value UP_BOUND 
    // Pre:    a_ < b_
    // Effect: contructs a B_spline_curve_2 object with control points 
    //         contr_pts of given degree and knots uniform distributed 
    //         over [a_,b_]
    B_spline_curve_2(const Points & contr_pts, unsigned degree = 3,
		     NumType a_=LOW_BOUND, 
		     NumType b_=UP_BOUND): 
                           S_curve(contr_pts), 		    
                           degree_(degree), knots_(),
	                   alg(&contr_points, &knots_, degree_),    	
                           b_x_FO(&alg), b_y_FO(&alg) {
	assert(a_ < b_);
	assert(contr_pts.size()>degree_);
	uniform_parametrisation(knots_, a_, b_, contr_pts.size());
	P_curve::init(&b_x_FO, &b_y_FO, a_, b_);	
	initial_approximation();
    };

    // B_spline_curve_2: B_spline_curve_2 -> B_spline_curve_2 
    // Parameters:
    //    bezier   B_spline curve
    // Effect: contructs a B_spline_curve_2 object which is a copy of 
    //         other   
    B_spline_curve_2(const B_spline_curve_2 & other) : 
    	S_curve(other.control_points()),
	degree_(other.degree()), knots_(other.get_knots()), 
	alg(&contr_points, &knots_, degree_),
	b_x_FO(&alg), b_y_FO(&alg) {	
	P_curve::init(other.get_x_func(), other.get_y_func(), 
		      other.get_lower_bound(), other.get_upper_bound());
	initial_approximation();
    };

    virtual ~B_spline_curve_2() {
    };

    // degree: void -> unsigned    
    // Returns the degree of this B-Spline
    unsigned degree() const {       
	return degree_; 
	
    };

    // set_degree: unsigned -> void    
    // Sets the degree of this B-Spline to d
    void set_degree(unsigned d) {
       	assert(d<contr_points.size());
	degree_= d; 
	alg.set_degree(d);       
    };

    // set_knots: ParamVector -> void    
    // Sets the knots of this B-Spline to d
    void set_knots(const ParamVector & knots) {       	
	knots_= knots;        
    };

    // get_name: void -> char*
    // Returns the string "B-Spline Curve"
    char * get_name() const {
	return "B-Spline Curve";
    };

    // get_knots(): void -> vector<NumType>
    // Returns the vector with the knots values for this spline
    ParamVector get_knots() const {
	return knots_;
    }

    // initial_approximation: unsigned -> Lin_approximant
    // Parameters:
    //     num       the number of segments in the linear approximant, 
    //               default value NUMBER  
    //     appr      the piecewise linear approximant of this curve
    // Pre: num > 1
    // Effect: initialize the piecewise linear approximant of 
    //         this b-spline to be the control polygon 
    void initial_approximation(unsigned num=NUMBER) {
      Lin_approximant tmp(contr_points.begin(), contr_points.end(), 
			  knots_.begin());
      lin_approx = tmp;
      eps = initial_epsilon();
    };    

    // derivative: unsigned -> B_spline_curve_2, bool
    // Papameters:    
    //    r          compute the r-th derivative of this B_spline 
    //               curve, default value r=1
    //    deriv      contains at the end the r-th derivative 
    //               (B_spline curve with n-r+1 control points), 
    //               if it exists  
    // Pre:     r>0 
    // Returns: true, if the r-th discrete derivative exists (r<n+1)
    //          false otherwise
    bool derivative(B_spline_curve_2 & deriv, unsigned r=1) const {
	assert(r > 0);
	if(degree_ < r) return false; 
	unsigned n = contr_points.size()-1;
	unsigned d = degree();
	Points q = control_points();
	ParamVector tmp_knots;
	//extend the knot vector with d/2 times knots[0] at the 
	//beginning and d/2 times knots[n] at the end
	int l = d/2;
	for(int i=0; i<l; i++) {
	   tmp_knots.push_back(knots_[0]); 
	}
	for(int i=0; i<=n; i++) {
	    tmp_knots.push_back(knots_[i]);
	}
	for(int i=0; i<l; i++) {
	   tmp_knots.push_back(knots_[n]); 
	}

	NumType denom;
	NumType coef = d;
	//the current evalueted derivative degree
	NumType r_current = 0;
	for(unsigned j = (n-1); j>=(n-r); j--) {
	    for(unsigned k = 0; k<=j; k++) {
		denom = tmp_knots[k+d] - tmp_knots[k+r_current];
		q[k] = CGAL::ORIGIN + (q[k+1]-q[k])*(coef/denom);
	    }
	    r_current++;
	    coef--;
	    //dieser break is noetig weil unsigned j=0-1=49...90 >(n-r)
	    //und die schleife geht ewig.
	    if(j==0) break;
	}

	PointList tmp(n-r+1);
	q.resize(n-r+1);
	copy(q.begin(), q.end(), tmp.begin());	
	deriv.set_control_points(tmp);
	deriv.set_degree(d-r);	
	uniform_parametrisation(tmp_knots, a, b, tmp.size());
	deriv.set_knots(tmp_knots);
	return true;
    }

    // evaluate_derivative_at: NumType, int -> Vector, bool
    // Papameters:
    //    t      the parameter at which the derivative is computed
    //    r      compute the r-th derivative of the curve, default r=1
    //    deriv   will contain the r-th derivative of the curve at t, if 
    //            it exists
    // Returns: true, if the r-th derivative at t exists
    //          false otherwise
    // Pre:  knots[0]<=t<=knots[size()-1]    
    // Overwrites Parametric_curve_2::evaluate_derivative_at()
    bool evaluate_derivative_at(NumType t, Vector & deriv, 
			     unsigned r=DEGREE) const {
	B_spline_curve_2 b_deriv;
	if(!derivative(b_deriv, r)) 
	    return false;
	assert(knots_[0]<=t && knots_[knots_.size()-1]>=t);
	assert(r>=1);	
	deriv = (b_deriv.calculate_point_at(t) - CGAL::ORIGIN);
	return true;
    }
    

    // subdivide: void -> void
    // Parameters:
    //    pts     contains the point sequence to be subdivided       
    // Effect: uses 1 round linear subdivision and (degree-1) rounds 
    //         averaging to compute finer point sequence pts 
    void subdivide(Points & pts) const {
	lin_subdivision(pts);
	//broia ma averaging triabva da e degree-1
	for(unsigned i=1; i<degree_; i++) {
	    averaging(pts);
	}
    }

    // sufficient_subdiv_level: NumType -> unsigned    
    // Parameter:
    //    epsilon   the tolerance, dafault value 0.001
    // Returns:  the number subdivision levels requiered to anssure that
    //           every segment of this curve will be approximated to within 
    //           given tolerance epsilon by the line segment of the
    //           refined control polygon
    unsigned sufficient_subdiv_level(NumType epsilon=EPSILON) const {
	if(S_curve::sufficient_subdiv_level(epsilon)==0) return 0;
	unsigned d = degree();
	assert(d>0);
	double M = max_second_diff_value();
	double arg = ((d+1)*M)/_to_float((24*epsilon));
	//log4(arg)=ln(arg)/ln(4)=ln(arg)/2*ln(2)=ln(arg)/2*M_LN2
	if(arg==0.0) return 0;
	unsigned res = (unsigned) ceil((log(arg))/(2*M_LN2));
	return res;
    }

    // rendering: NumType -> Points
    // Parameters:
    //   epsilon   the tolerance
    //   pts       vector of points wich approximates this curve within 
    //             a tolerance epsilon
    // Performs the needed rounds subdivision to refine the control polygon 
    // The refined control polygon approximates the curve to whitin the
    // given tolerance epsilon
    void rendering(Points& pts, NumType epsilon=EPSILON) const {
	pts = contr_points;
	unsigned level = sufficient_subdiv_level(epsilon);
	for(unsigned i=0; i<=level; i++) {
	    subdivide(pts);
	}
    }

    // refine_approximantion: NumType -> void
    // Parameters:
    //    epsilon    for all t in [a,b] dist(appr(t), curve(t))<=epsilon,
    //               default value EPSILON
    // Effect: refines the piecewise linear approximant of this curve so that
    //         for all t in [a,b] dist(appr(t), curve(t))<=epsilon
    // Overwrites P_curve::refine_approximation()
    void refine_approximation(NumType epsilon=EPSILON, 
			      bool use_deriv=true) const {
	Points pts;
	rendering(pts, epsilon);
	ParamVector param;
	uniform_parametrisation(param, a, b, pts.size());
	lin_approx.set_points(pts.begin(), pts.end(), param.begin());
	eps = epsilon;
    }

    // update: PointList -> SegmentList
    // Parameters:
    //    in     container with the control points of this b-spline
    //    out    container with the segments for drawing this b-spline curve
    // Effect: calls by every change of the input scene (the control points)
    //         and calculates the segments of the output scene (the curve)
    // Overwrites geowin_update<PointList, SegmentList>::update()
    void update (const PointList& in, SegmentList& out) {	
	out.clear();
	if(in.size() < 2) return;				
	set_control_points(in);	       			
	Points pts;
	rendering(pts);
	for(unsigned i = 1; i<pts.size(); i++) {
	    Segment s(pts[i-1], pts[i]);
	    out.push_back(s);
	}		
    }
    
    // draw: void -> void 
    // Effect: opens a GeoWin window and draws this b-spline curve
    // Overwrites Parametric_curve_2::draw()
    void draw() {
	geowin_init_default_type((PointList*)0, 
				 leda_string("ContrPointList"));
	geowin_init_default_type((SegmentList*)0, 
				 leda_string("SegmentList"));	
	GeoWin GW("B-Spline Curves Demo");
	GW.set_action( A_LEFT | A_DRAG | A_OBJECT, my_geo_object_dragging);
	
	PointList in_list;	
	in_list.resize(contr_points.size());
	copy(contr_points.begin(), contr_points.end(), in_list.begin());

	//the input scene is the scene with the control points
	geo_scene in_scene = GW.new_scene(in_list);
	GW.set_point_style(in_scene, leda_disc_point);
	GW.set_color(in_scene, leda_red);
	GW.set_visible(in_scene, true);
	GW.activate(in_scene);

	//result scene for the control polygon	
	geo_contr_polygon contr_polygon;
	geo_scene polygon_res_scene = GW.new_scene(contr_polygon, 
		          in_scene, leda_string("Control Polygon"));
	GW.set_color(polygon_res_scene, leda_red);
	GW.set_line_width(polygon_res_scene, 1);
	GW.set_visible(polygon_res_scene, true);	
	
	P_curve::draw(GW, in_scene, false);
	GW.edit(in_scene); 
    } 

    // bbox: void -> Bbox    
    // Returns: the bounding box of this b-spline curve wich is the 
    //          bbox of the control points
    // Overwrites P_curve::bbox()
    Bbox bbox() const {	
	Lin_approximant tmp(contr_points.begin(), contr_points.end(), 
			    a, b);
	return tmp.bbox();	  
    };

    // has_on: Point, NumType -> bool
    // Parameters: 
    //     pt       the point is tested if it is on the curve
    //     epsilon  the upper bound for the distance from pt to the 
    //              approximation of the curve
    // Returns: true, if dist(pt, lin_approx)<=epsilon. This means that
    //          the point pt is at most 2*epsilon away from the curve.
    //          false otherwise
    // Tests approximativ if the point pt lies on the curve.
    // This function just solves the ambiguous problem with P_curve::has_on()
    // and S_curve::has_on() (B-Splines inherits from both)
    bool has_on(const Point& pt, NumType epsilon=EPSILON) const {
	return P_curve::has_on(pt, epsilon);
    }

protected:
    //degree of the B-Spline, the degree of continuity is degree_-1
    unsigned degree_;
    ParamVector knots_;

private:
    //auxilary variables for the contruction
    A alg;
    B_spline_x b_x_FO;
    B_spline_y b_y_FO;    

};
#endif //CGAL_B_SPLINE_CURVE_2_H
