#ifndef CGAL_HORNER_FO_H
#define CGAL_HORNER_FO_H

#include "Types.h"
#include "Curve_constants.h"
#include "Helpers.h"

// Template function object class which implements the Horner algorithm
// for calculating a point on a Bezier curve at given parameter value t
// Parameters:
//    NumType_    the number type of the points coordinates 
template <class NumType_> 
class Horner_FO {
public:
    typedef NumType_ NumType;
    typedef CGAL::Cartesian<NumType> REP;
    typedef CGAL::Point_2<REP> Point;   
    typedef vector<Point> Points;
    typedef CGAL::Vector_2<REP> Vector;
    typedef vector<Vector> Vectors;

    Horner_FO() : contr_pts(NULL), a(LOW_BOUND), b(UP_BOUND) {
    }
    
    Horner_FO(Points * contr_points, NumType a_=LOW_BOUND, 
	      NumType b_=UP_BOUND) : 
	contr_pts(contr_points), a(a_), b(b_) {
    }    

    // operator(): NumType -> Point
    // Parameters:
    //   t    the parameter value
    // Pre: a<=t<=b
    //      there is at least one control point
    // Returns: a point on the Bezier curve calculated with the Horner
    //          algorithm at t
    Point operator() (NumType t) const {
	assert(!(*contr_pts).empty());
	assert(a<=t && t<=b);
	int n = (*contr_pts).size()-1;	
	if(t==a) return (* contr_pts)[0];	
	if(t==b) return (* contr_pts)[n];
	NumType c[n + 1];
	binomials(c, n);	
	Vectors tmp_v(n+1);
	for(int i=0; i<=n; i++) {
	    tmp_v[i] = (* contr_pts)[i] - CGAL::ORIGIN;
	}
	NumType tn, dtn;
	Vector tmp;
	if(t<=((b+a)/2.0)) {
	    tn = power(((b-t)/(b-a)), n);
	    dtn = (t-a)/(b-t);
	    tmp = tmp_v[n];
	    for(int i=n; i>=1; i--) {
		tmp = tmp*dtn + tmp_v[i-1]*c[i-1];	      
	    }	    
	}
	else {
	    tn = power((t-a)/(b-a), n);
	    dtn =  (b-t)/(t-a);
	    tmp = tmp_v[0];
	    for(int i=1; i<=n; i++) {
		tmp = tmp*dtn + tmp_v[i]*c[i];	      
	    }
	}
	return (CGAL::ORIGIN + tmp*tn);	
    }
    
private:
    Points * contr_pts;
    NumType a, b;
};
#endif //CGAL_HORNER_FO_H
